/* ////////////////////////////////////////////////////////////

  File Name: ysshellextedit.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "ysshellextedit.h"
#include "ysshellextio.h"


class YsShellExtEdit::EditLog
{
public:
	YSBOOL isRememberSelection; 
	// Remember Selection is a special undo type.
	// If undoPtr or redoPtr stops at Remember-Selection node, it will go one
	// more step to execute remember selection.

	YSSIZE_T undoCtr;
	EditLog *prev,*next;

	EditLog();
	virtual ~EditLog(){};
	virtual void Undo(YsShellExtEdit *shl)=0;
	virtual void Redo(YsShellExtEdit *shl)=0;
	virtual void NeverBeUndoneAgain(YsShellExtEdit *){};
	virtual void NeverBeRedoneAgain(YsShellExtEdit *){};
};

YsShellExtEdit::EditLog::EditLog()
{
	isRememberSelection=YSFALSE;
	undoCtr=0;
	prev=NULL;
	next=NULL;
}


////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoRememberSelection : public YsShellExtEdit::EditLog
{
public:
	YsArray <unsigned int> selectedVtKey,selectedPlKey,selectedCeKey,selectedFgKey,selectedVlKey;

	UndoRememberSelection();

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	void SetSelection(YsShellExtEdit *);
	void CaptureSelection(const YsShellExtEdit *);
};

YsShellExtEdit::UndoRememberSelection::UndoRememberSelection()
{
	isRememberSelection=YSTRUE;
}

void YsShellExtEdit::UndoRememberSelection::Undo(YsShellExtEdit *shl)
{
	SetSelection(shl);
}

void YsShellExtEdit::UndoRememberSelection::Redo(YsShellExtEdit *shl)
{
	SetSelection(shl);
}

void YsShellExtEdit::UndoRememberSelection::SetSelection(YsShellExtEdit *shl)
{
	shl->selectedVtKey=selectedVtKey;
	shl->selectedPlKey=selectedPlKey;
	shl->selectedCeKey=selectedCeKey;
	shl->selectedFgKey=selectedFgKey;
	shl->selectedVlKey=selectedVlKey;
}

void YsShellExtEdit::UndoRememberSelection::CaptureSelection(const YsShellExtEdit *shl)
{
	selectedVtKey=shl->selectedVtKey;
	selectedPlKey=shl->selectedPlKey;
	selectedCeKey=shl->selectedCeKey;
	selectedFgKey=shl->selectedFgKey;
	selectedVlKey=shl->selectedVlKey;
}

////////////////////////////////////////////////////////////

YsShellExtEdit::StopIncUndo::StopIncUndo(YsShellExtEdit *shl)
{
	this->shl=shl;
	this->incUndo=shl->PushStopIncUndo();
}

YsShellExtEdit::StopIncUndo::StopIncUndo(YsShellExtEdit &shl)
{
	this->shl=&shl;
	this->incUndo=shl.PushStopIncUndo();
}

YsShellExtEdit::StopIncUndo::~StopIncUndo()
{
	shl->PopIncUndo(incUndo);
}

////////////////////////////////////////////////////////////

YSBOOL YsShellExtEdit::IsEmptyAndHasNeverBeenEdited(void) const
{
	if(0==GetNumVertex() &&
	   0==GetNumPolygon() &&
	   0==GetNumConstEdge() &&
	   0==GetNumFaceGroup() &&
	   0==GetNumVolume() &&
	   NULL==undoPtr &&
	   NULL==redoPtr)
	{
		return YSTRUE;
	}
	return YSFALSE;
}

void YsShellExtEdit::CleanUpEditLog(void)
{
	CleanUpUndoLog();
	CleanUpRedoLog();
}

void YsShellExtEdit::CleanUpUndoLog(void)
{
	YsArray <EditLog *> toDel;
	for(EditLog *ptr=undoPtr; NULL!=ptr; ptr=ptr->prev)
	{
		toDel.Append(ptr);
	}
	for(YSSIZE_T idx=0; idx<toDel.GetN(); ++idx)
	{
		toDel[idx]->NeverBeUndoneAgain(this);
		delete toDel[idx];
	}
	undoPtr=NULL;
	if(NULL!=redoPtr)
	{
		redoPtr->prev=NULL;
	}
}

void YsShellExtEdit::CleanUpRedoLog(void)
{
	YsArray <EditLog *> toDel;
	for(EditLog *ptr=redoPtr; NULL!=ptr; ptr=ptr->next)
	{
		toDel.Append(ptr);
	}
	toDel.Invert();  // Newer entity may be using an older entity.
	for(YSSIZE_T idx=0; idx<toDel.GetN(); ++idx)
	{
		toDel[idx]->NeverBeRedoneAgain(this);
		delete toDel[idx];
	}
	redoPtr=NULL;
	if(NULL!=undoPtr)
	{
		undoPtr->next=NULL;
	}
}

void YsShellExtEdit::AddUndo(EditLog *editLog)
{
	if(YSTRUE==takeEditLog)
	{
		CleanUpRedoLog();

		if(NULL==undoPtr || undoPtr->undoCtr!=undoCtr)
		{
			InsertRememberSelection();
		}

		editLog->undoCtr=undoCtr;
		undoCtr+=incUndo;

		if(NULL!=undoPtr)
		{
			undoPtr->next=editLog;
			editLog->prev=undoPtr;
		}

		undoPtr=editLog;
		redoPtr=NULL;
	}
	else
	{
		printf("%s %d\n",__FUNCTION__,__LINE__);
		printf("  This function must not be called when takeEditLog is YSFALSE.\n");
		delete editLog;
	}
}

void YsShellExtEdit::InsertRememberSelection(void)
{
	UndoRememberSelection *undo=new UndoRememberSelection;
	undo->CaptureSelection(this);
	undo->undoCtr+=undoCtr;
	undoCtr+=incUndo;
	undo->isRememberSelection=YSTRUE;

	if(NULL!=undoPtr)
	{
		undoPtr->next=undo;
		undo->prev=undoPtr;
	}
	undoPtr=undo;
	redoPtr=NULL;
}

YsShellExtEdit::YsShellExtEdit()
{
	takeEditLog=YSTRUE;
	undoCtr=0;
	incUndo=1;
	undoPtr=NULL;
	redoPtr=NULL;

	lastSavedUndoCtr=0;
	lastSavedUndoPtr=NULL;

	YsShellExt::EnableSearch();
}

YsShellExtEdit::~YsShellExtEdit()
{
	YsArray <EditLog *> toDel;
	for(EditLog *ptr=undoPtr; NULL!=ptr; ptr=ptr->prev)
	{
		toDel.Append(ptr);
	}
	for(EditLog *ptr=redoPtr; NULL!=ptr; ptr=ptr->next)
	{
		toDel.Append(ptr);
	}
	for(YSSIZE_T idx=0; idx<toDel.GetN(); ++idx)
	{
		delete toDel[idx];
	}
}

YsShellExtEdit::operator const YsShell &() const
{
	return *this;
}

YsShellExtEdit::operator const YsShellExt &() const
{
	return *this;
}

void YsShellExtEdit::CleanUp(void)
{
printf("%s %d\n",__FUNCTION__,__LINE__);
printf("Not implemented yet.\n");
exit(1);

	selectedVtKey.CleanUp();
	selectedPlKey.CleanUp();
	selectedCeKey.CleanUp();
}

YSRESULT YsShellExtEdit::LoadSrf(YsTextInputStream &inStream)
{
	YsShellExtReader srfReader;
	return srfReader.MergeSrf(*this,inStream);
}

YSRESULT YsShellExtEdit::LoadObj(YsTextInputStream &inStream)
{
	YsShellExtObjReader::ReadOption opt;
	YsShellExtObjReader objReader;
	return objReader.ReadObj(*this,inStream,opt);
}

YSRESULT YsShellExtEdit::LoadStl(const char fn[])
{
	return YsShellExt::LoadStl(fn);
}

YSBOOL YsShellExtEdit::IsModified(void) const
{
	EditLog *effectiveCurrentUndoPtr,*effectiveLastSavedUndoPtr;

	for(effectiveCurrentUndoPtr=undoPtr; 
	    NULL!=effectiveCurrentUndoPtr && YSTRUE==effectiveCurrentUndoPtr->isRememberSelection; 
	    effectiveCurrentUndoPtr=effectiveCurrentUndoPtr->prev);

	for(effectiveLastSavedUndoPtr=lastSavedUndoPtr;
	    NULL!=effectiveLastSavedUndoPtr && YSTRUE==effectiveLastSavedUndoPtr->isRememberSelection;
	    effectiveLastSavedUndoPtr=effectiveLastSavedUndoPtr->prev);

	if(NULL==effectiveCurrentUndoPtr && NULL==effectiveLastSavedUndoPtr)
	{
		return YSFALSE;
	}
	else if(NULL!=effectiveCurrentUndoPtr &&
	        effectiveCurrentUndoPtr==effectiveLastSavedUndoPtr &&
	        lastSavedUndoCtr==effectiveCurrentUndoPtr->undoCtr)
	{
		return YSFALSE;
	}
	return YSTRUE;
}

YsShellExtEdit::SavePoint YsShellExtEdit::GetSavePoint(void) const
{
	SavePoint savePoint;
	if(NULL!=undoPtr)
	{
		savePoint.undoCtr=undoPtr->undoCtr;
	}
	else
	{
		savePoint.undoCtr=0;
	}
	savePoint.undoPtr=undoPtr;
	return savePoint;
}

YSBOOL YsShellExtEdit::IsModified(const YsShellExtEdit::SavePoint &savePoint) const
{
	const EditLog *effectiveCurrentUndoPtr,*effectiveLastSavedUndoPtr;

	for(effectiveCurrentUndoPtr=undoPtr; 
	    NULL!=effectiveCurrentUndoPtr && YSTRUE==effectiveCurrentUndoPtr->isRememberSelection; 
	    effectiveCurrentUndoPtr=effectiveCurrentUndoPtr->prev);

	for(effectiveLastSavedUndoPtr=savePoint.undoPtr;
	    NULL!=effectiveLastSavedUndoPtr && YSTRUE==effectiveLastSavedUndoPtr->isRememberSelection;
	    effectiveLastSavedUndoPtr=effectiveLastSavedUndoPtr->prev);

	if(NULL==effectiveCurrentUndoPtr && NULL==effectiveLastSavedUndoPtr)
	{
		return YSFALSE;
	}
	else if(NULL!=effectiveCurrentUndoPtr &&
	        effectiveCurrentUndoPtr==effectiveLastSavedUndoPtr &&
	        savePoint.undoCtr==effectiveCurrentUndoPtr->undoCtr)
	{
		return YSFALSE;
	}
	return YSTRUE;
}

void YsShellExtEdit::Saved(void) const
{
	EditLog *effectiveCurrentUndoPtr;
	for(effectiveCurrentUndoPtr=undoPtr; 
	    NULL!=effectiveCurrentUndoPtr && YSTRUE==effectiveCurrentUndoPtr->isRememberSelection; 
	    effectiveCurrentUndoPtr=effectiveCurrentUndoPtr->prev);

	lastSavedUndoPtr=effectiveCurrentUndoPtr;
	if(NULL!=lastSavedUndoPtr)
	{
		lastSavedUndoCtr=lastSavedUndoPtr->undoCtr;
	}
	else
	{
		lastSavedUndoCtr=0;
	}
}

void YsShellExtEdit::SetFileName(const wchar_t fn[]) const
{
	this->fn.Set(fn);
}

void YsShellExtEdit::SetFileName(const char fn[]) const
{
	this->fn.SetUTF8String(fn);
}

const wchar_t *YsShellExtEdit::GetFileName(void) const
{
	return fn;
}

const wchar_t *YsShellExtEdit::GetFileName(YsWString &fn) const
{
	fn=this->fn;
	return fn;
}

const char *YsShellExtEdit::GetFileName(YsString &fn) const
{
	fn.EncodeUTF8 <wchar_t> (this->fn);
	return fn;
}

YSRESULT YsShellExtEdit::Undo(void)
{
	if(NULL!=undoPtr && YSTRUE==undoPtr->isRememberSelection)
	{
		SkipUndoOneStep();  // Just skip it.
	}

	if(NULL!=undoPtr)
	{
		const YSSIZE_T undoCtr=undoPtr->undoCtr;
		while(NULL!=undoPtr && undoCtr==undoPtr->undoCtr)
		{
			UndoOneStep();
		}

		if(NULL!=redoPtr && YSTRUE!=redoPtr->isRememberSelection &&
		   NULL!=undoPtr && YSTRUE==undoPtr->isRememberSelection)
		{
			UndoOneStep();
		}

		return YSOK;
	}
	return YSERR;
}

void YsShellExtEdit::UndoOneStep(void)
{
	undoPtr->Undo(this);
	redoPtr=undoPtr;
	undoPtr=undoPtr->prev;
}

void YsShellExtEdit::SkipUndoOneStep(void)
{
	redoPtr=undoPtr;
	undoPtr=undoPtr->prev;
}

YSRESULT YsShellExtEdit::Redo(void)
{
	if(NULL!=redoPtr && YSTRUE==redoPtr->isRememberSelection)
	{
		SkipRedoOneStep(); // Just skip it.
	}

	if(NULL!=redoPtr)
	{
		const YSSIZE_T undoCtr=redoPtr->undoCtr;
		while(NULL!=redoPtr && undoCtr==redoPtr->undoCtr)
		{
			RedoOneStep();
		}

		if(NULL!=undoPtr && YSTRUE!=undoPtr->isRememberSelection &&
		   NULL!=redoPtr && YSTRUE==redoPtr->isRememberSelection)
		{
			RedoOneStep();
		}

		return YSOK;
	}
	return YSERR;
}

void YsShellExtEdit::RedoOneStep(void)
{
	redoPtr->Redo(this);
	undoPtr=redoPtr;
	redoPtr=redoPtr->next;
}

void YsShellExtEdit::SkipRedoOneStep(void)
{
	undoPtr=redoPtr;
	redoPtr=redoPtr->next;
}

int YsShellExtEdit::PushStopIncUndo(void)
{
	int prevIncUndo=incUndo;
	incUndo=0;
	return prevIncUndo;
}

void YsShellExtEdit::PopIncUndo(int incUndoIn)
{
	undoCtr+=incUndoIn;
	this->incUndo=incUndoIn;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoAddVertex : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellVertexHandle> vtHdArray;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *shl);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoAddVertex::Undo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=vtHdArray.GetN()-1; 0<=idx; --idx)
	{
		shl->FreezeVertex(vtHdArray[idx]);
	}
}

void YsShellExtEdit::UndoAddVertex::Redo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN(); ++idx)
	{
		shl->MeltVertex(vtHdArray[idx]);
	}
}

void YsShellExtEdit::UndoAddVertex::NeverBeUndoneAgain(YsShellExtEdit *)
{
}

void YsShellExtEdit::UndoAddVertex::NeverBeRedoneAgain(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN(); ++idx)
	{
		YsShellExt *shlExt=shl;
		shlExt->DeleteFrozenVertex(vtHdArray[idx]);
	}
}

YsShellVertexHandle YsShellExtEdit::AddVertex(const YsVec3 &pos)
{
	YsShellExtEdit::UndoAddVertex *undo=NULL;

	if(YSTRUE==takeEditLog)
	{
		undo=new YsShellExtEdit::UndoAddVertex;
	}

	YsShellVertexHandle vtHd=YsShellExt::AddVertex(pos);

	if(NULL!=undo)
	{
		undo->vtHdArray.MakeUnitLength(vtHd);
		AddUndo(undo);
	}

	return vtHd;
}

YSRESULT YsShellExtEdit::AddMultiVertex(YsArray <YsShellVertexHandle> &vtHdArray,const YSSIZE_T nVt,const YsVec3 pos[])
{
	YsShellExtEdit::UndoAddVertex *undo=NULL;

	if(YSTRUE==takeEditLog)
	{
		undo=new YsShellExtEdit::UndoAddVertex;
	}

	vtHdArray.Set(nVt,NULL);
	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		vtHdArray[idx]=YsShellExt::AddVertex(pos[idx]);
	}

	if(NULL!=undo)
	{
		undo->vtHdArray=vtHdArray;
		AddUndo(undo);
	}

	return YSOK;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoSetVertexPosition : public YsShellExtEdit::EditLog
{
public:
	YsShellVertexHandle vtHd;
	YsVec3 prvPos,newPos;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoSetVertexPosition::Undo(YsShellExtEdit *shl)
{
	YsShell *rawShl=(YsShell *)shl;
	rawShl->ModifyVertexPosition(vtHd,prvPos);
}

void YsShellExtEdit::UndoSetVertexPosition::Redo(YsShellExtEdit *shl)
{
	YsShell *rawShl=(YsShell *)shl;
	rawShl->ModifyVertexPosition(vtHd,newPos);
}

YSRESULT YsShellExtEdit::SetVertexPosition(YsShellVertexHandle vtHd,const YsVec3 &newPos)
{
	UndoSetVertexPosition *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new UndoSetVertexPosition;
		undo->vtHd=vtHd;
		GetVertexPosition(undo->prvPos,vtHd);
		undo->newPos=newPos;
		AddUndo(undo);
	}

	YsShell::ModifyVertexPosition(vtHd,newPos);

	return YSOK;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoSetMultiVertexPosition : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellVertexHandle> vtHdArray;
	YsArray <YsVec3> prvPosArray,newPosArray;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoSetMultiVertexPosition::Undo(YsShellExtEdit *shl)
{
	YsShell *rawShl=(YsShell *)shl;
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN() && idx<prvPosArray.GetN(); ++idx)
	{
		rawShl->ModifyVertexPosition(vtHdArray[idx],prvPosArray[idx]);
	}
}

void YsShellExtEdit::UndoSetMultiVertexPosition::Redo(YsShellExtEdit *shl)
{
	YsShell *rawShl=(YsShell *)shl;
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN() && idx<newPosArray.GetN(); ++idx)
	{
		rawShl->ModifyVertexPosition(vtHdArray[idx],newPosArray[idx]);
	}
}

YSRESULT YsShellExtEdit::SetMultiVertexPosition(int nVtx,const YsShellVertexHandle vtHdArray[],const YsVec3 vtxPosArray[])
{
	UndoSetMultiVertexPosition *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new UndoSetMultiVertexPosition;
		undo->vtHdArray.Set(nVtx,vtHdArray);
		undo->prvPosArray.Set(nVtx,NULL);
		undo->newPosArray.Set(nVtx,NULL);
		for(YSSIZE_T idx=0; idx<nVtx; ++idx)
		{
			GetVertexPosition(undo->prvPosArray[idx],vtHdArray[idx]);
			undo->newPosArray[idx]=vtxPosArray[idx];
		}
		AddUndo(undo);
	}

	for(YSSIZE_T idx=0; idx<nVtx; ++idx)
	{
		YsShell::ModifyVertexPosition(vtHdArray[idx],vtxPosArray[idx]);
	}

	return YSOK;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoVertexAttribChange : public YsShellExtEdit::EditLog
{
public:
	YsShellVertexHandle vtHd;
	VertexAttrib prvAttrib,newAttrib;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoVertexAttribChange::Undo(YsShellExtEdit *shl)
{
	VertexAttrib *attr=shl->GetVertexAttrib(vtHd);
	attr->CopyFrom(prvAttrib);
}

void YsShellExtEdit::UndoVertexAttribChange::Redo(YsShellExtEdit *shl)
{
	VertexAttrib *attr=shl->GetVertexAttrib(vtHd);
	attr->CopyFrom(newAttrib);
}

// YSRESULT YsShellExtEdit::SetVertexSomeAttrib(YsShellVertexHandle vtHd,new attrib)
// {
// 	VertexAttrib *attr=GetVertexAttrib(vtHd);
// 	if(NULL!=attr)
// 	{
// 		UndoVertexAttribChange *undo=NULL;
// 		if(YSTRUE==takeEditLog)
// 		{
// 			undo=new UndoVertexAttribChange;
// 			undo->vtHd=vtHd;
// 			undo->prvAttrib.CopyFrom(*attr);
// 		}
// 
// 		attr->?=?;
// 
// 		if(NULL!=undo)
// 		{
// 			undo->newAttrib.CopyFrom(*attr);
// 			AddUndo(undo);
// 		}
// 		return YSOK;
// 	}
// 	return YSERR;
// }

YSRESULT YsShellExtEdit::SetVertexRoundFlag(YsShellVertexHandle vtHd,YSBOOL round)
{
	VertexAttrib *attr=GetVertexAttrib(vtHd);
	if(NULL!=attr)
	{
		UndoVertexAttribChange *undo=NULL;
		if(YSTRUE==takeEditLog)
		{
			undo=new UndoVertexAttribChange;
			undo->vtHd=vtHd;
			undo->prvAttrib.CopyFrom(*attr);
		}

		attr->SetRound(round);

		if(NULL!=undo)
		{
			undo->newAttrib.CopyFrom(*attr);
			AddUndo(undo);
		}
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExtEdit::SetVertexConstrainedFlag(YsShellVertexHandle vtHd,YSBOOL constFlag)
{
	VertexAttrib *attr=GetVertexAttrib(vtHd);
	if(NULL!=attr)
	{
		UndoVertexAttribChange *undo=NULL;
		if(YSTRUE==takeEditLog)
		{
			undo=new UndoVertexAttribChange;
			undo->vtHd=vtHd;
			undo->prvAttrib.CopyFrom(*attr);
		}

		attr->SetConstrained(constFlag);

		if(NULL!=undo)
		{
			undo->newAttrib.CopyFrom(*attr);
			AddUndo(undo);
		}
		return YSOK;
	}
	return YSERR;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoDeleteVertex : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellVertexHandle> vtHdArray;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *shl);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoDeleteVertex::Undo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN(); ++idx)
	{
		shl->MeltVertex(vtHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeleteVertex::Redo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=vtHdArray.GetN()-1; 0<=idx; --idx)
	{
		shl->FreezeVertex(vtHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeleteVertex::NeverBeUndoneAgain(YsShellExtEdit *shl)
{
	// Means, the vertices will never be back alive again.  Can permanently be deleted.
	for(YSSIZE_T idx=0; idx<vtHdArray.GetN(); ++idx)
	{
		YsShellExt *shlExt=shl;
		shlExt->DeleteFrozenVertex(vtHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeleteVertex::NeverBeRedoneAgain(YsShellExtEdit *)
{
}

YSRESULT YsShellExtEdit::DeleteVertex(YsShellVertexHandle vtHd)
{
	const YsShellVertexHandle vtHdArray[1]={vtHd};
	return DeleteMultiVertex(1,vtHdArray);
}

YSRESULT YsShellExtEdit::DeleteMultiVertex(const YSSIZE_T nVt,const YsShellVertexHandle vtHd[])
{
	YSRESULT res=YSOK;

	YsArray <YsShellVertexHandle> actuallyDeleted;
	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		if(YSOK==YsShellExt::FreezeVertex(vtHd[idx]))
		{
			actuallyDeleted.Append(vtHd[idx]);
		}
		else
		{
			res=YSERR;
		}
	}

	if(YSTRUE==takeEditLog)
	{
		YsShellExtEdit::UndoDeleteVertex *undo=new YsShellExtEdit::UndoDeleteVertex;
		undo->vtHdArray=actuallyDeleted;
		AddUndo(undo);
	}

	return res;
}

////////////////////////////////////////////////////////////

YSRESULT YsShellExtEdit::SelectVertex(const YSSIZE_T nVt,const YsShellVertexHandle vtHd[])
{
	selectedVtKey.Set(nVt,NULL);
	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		selectedVtKey[idx]=GetSearchKey(vtHd[idx]);
	}
	return YSOK;
}

YSRESULT YsShellExtEdit::AddSelectedVertex(const YSSIZE_T nVt,const YsShellVertexHandle vtHd[])
{
	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		selectedVtKey.Append(GetSearchKey(vtHd[idx]));
	}
	return YSOK;
}

YsArray <YsShellVertexHandle> YsShellExtEdit::GetSelectedVertex(void) const
{
	YsArray <YsShellVertexHandle> selVtHd;
	GetSelectedVertex(selVtHd);
	return selVtHd;
}

YSRESULT YsShellExtEdit::SelectPolygon(const YSSIZE_T nPl,const YsShellPolygonHandle plHd[])
{
	selectedPlKey.Set(nPl,NULL);
	for(YSSIZE_T idx=0; idx<nPl; ++idx)
	{
		selectedPlKey[idx]=GetSearchKey(plHd[idx]);
	}
	return YSOK;
}

YSRESULT YsShellExtEdit::AddSelectedPolygon(const YSSIZE_T nPl,const YsShellPolygonHandle plHd[])
{
	for(YSSIZE_T idx=0; idx<nPl; ++idx)
	{
		selectedPlKey.Append(GetSearchKey(plHd[idx]));
	}
	return YSOK;
}

YsArray <YsShellPolygonHandle> YsShellExtEdit::GetSelectedPolygon(void) const
{
	YsArray <YsShellPolygonHandle> selPlHd;
	GetSelectedPolygon(selPlHd);
	return selPlHd;
}

YSRESULT YsShellExtEdit::SelectConstEdge(const YSSIZE_T nVt,const YsShellExt::ConstEdgeHandle ceHd[])
{
	selectedCeKey.Set(nVt,NULL);
	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		selectedCeKey[idx]=GetSearchKey(ceHd[idx]);
	}
	return YSOK;
}

YSRESULT YsShellExtEdit::AddSelectedConstEdge(const YSSIZE_T nVt,const YsShellExt::ConstEdgeHandle ceHd[])
{
	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		selectedCeKey.Append(GetSearchKey(ceHd[idx]));
	}
	return YSOK;
}

YsArray <YsShellExt::ConstEdgeHandle> YsShellExtEdit::GetSelectedConstEdge(void) const
{
	YsArray <YsShellExt::ConstEdgeHandle> selCeHd;
	GetSelectedConstEdge(selCeHd);
	return selCeHd;
}

YSRESULT YsShellExtEdit::SelectFaceGroup(const YSSIZE_T nFg,const YsShellExt::FaceGroupHandle fgHd[])
{
	selectedFgKey.Set(nFg,NULL);
	for(YSSIZE_T idx=0; idx<nFg; ++idx)
	{
		selectedFgKey[idx]=GetSearchKey(fgHd[idx]);
	}
	return YSOK;
}

YSRESULT YsShellExtEdit::AddSelectedFaceGroup(const YSSIZE_T nFg,const YsShellExt::FaceGroupHandle fgHd[])
{
	for(YSSIZE_T idx=0; idx<nFg; ++idx)
	{
		selectedFgKey.Append(GetSearchKey(fgHd[idx]));
	}
	return YSOK;
}

YsArray <YsShellExt::FaceGroupHandle> YsShellExtEdit::GetSelectedFaceGroup(void) const
{
	YsArray <YsShellExt::FaceGroupHandle> selFgHd;
	GetSelectedFaceGroup(selFgHd);
	return selFgHd;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoAddPolygon : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellPolygonHandle> plHdArray;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *shl);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoAddPolygon::Undo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=plHdArray.GetN()-1; 0<=idx; --idx)
	{
		YsShellExt *shlExt=shl;
		shlExt->FreezePolygon(plHdArray[idx]);
	}
}

void YsShellExtEdit::UndoAddPolygon::Redo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		YsShellExt *shlExt=shl;
		shlExt->MeltPolygon(plHdArray[idx]);
	}
}

void YsShellExtEdit::UndoAddPolygon::NeverBeUndoneAgain(YsShellExtEdit *)
{
}

void YsShellExtEdit::UndoAddPolygon::NeverBeRedoneAgain(YsShellExtEdit *shl)
{
	// Means that the polygon will never be restored again.  Can be permanently deleted.
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		YsShellExt *shlExt=shl;
		shlExt->DeleteFrozenPolygon(plHdArray[idx]);
	}
}

YsShellPolygonHandle YsShellExtEdit::AddPolygon(YSSIZE_T nVt,const YsShellVertexHandle vtHdArray[])
{
	YsShellExtEdit::UndoAddPolygon *undo=NULL;

	if(YSTRUE==takeEditLog)
	{
		undo=new YsShellExtEdit::UndoAddPolygon;
	}

	YsShellPolygonHandle plHd=YsShellExt::AddPolygon(nVt,vtHdArray);

	if(NULL!=undo)
	{
		undo->plHdArray.MakeUnitLength(plHd);
		AddUndo(undo);
	}

	return plHd;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoDeletePolygon : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellPolygonHandle> plHdArray;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *shl);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoDeletePolygon::Undo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		shl->MeltPolygon(plHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeletePolygon::Redo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		shl->FreezePolygon(plHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeletePolygon::NeverBeUndoneAgain(YsShellExtEdit *shl)
{
	// Won't be alive again.
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		YsShellExt *shlExt=shl;
		shlExt->DeleteFrozenPolygon(plHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeletePolygon::NeverBeRedoneAgain(YsShellExtEdit *)
{
}

YSRESULT YsShellExtEdit::DeletePolygon(YsShellPolygonHandle plHd)
{
	const YsShellPolygonHandle plHdArray[1]={plHd};
	return DeleteMultiPolygon(1,plHdArray);
}

YSRESULT YsShellExtEdit::DeleteMultiPolygon(YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[])
{
	YSRESULT res=YSOK;

	YsArray <YsShellPolygonHandle> actuallyDeleted;
	for(YSSIZE_T idx=0; idx<nPl; ++idx)
	{
		if(YSOK==FreezePolygon(plHdArray[idx]))
		{
			actuallyDeleted.Append(plHdArray[idx]);
		}
		else
		{
			res=YSERR;
		}
	}

	if(YSTRUE==takeEditLog && 0<actuallyDeleted.GetN())
	{
		YsShellExtEdit::UndoDeletePolygon *undo=new YsShellExtEdit::UndoDeletePolygon;
		undo->plHdArray=actuallyDeleted;
		AddUndo(undo);
	}

	return res;
}

////////////////////////////////////////////////////////////

YSRESULT YsShellExtEdit::ForceDeletePolygon(YsShellPolygonHandle plHd)
{
	DeleteFaceGroupPolygon(plHd);
	return DeletePolygon(plHd);
}

YSRESULT YsShellExtEdit::ForceDeleteMultiPolygon(YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[])
{
	for(YSSIZE_T idx=0; idx<nPl; ++idx)
	{
		DeleteFaceGroupPolygon(plHdArray[idx]);
	}
	return DeleteMultiPolygon(nPl,plHdArray);
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoSetPolygonVertex : public YsShellExtEdit::EditLog
{
public:
	YsShellPolygonHandle plHd;
	YsArray <YsShellVertexHandle> prvVtHd,newVtHd;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoSetPolygonVertex::Undo(YsShellExtEdit *shl)
{
	YsShellExt *rawShl=(YsShellExt *)shl;
	rawShl->SetPolygonVertex(plHd,prvVtHd.GetN(),prvVtHd);
}

void YsShellExtEdit::UndoSetPolygonVertex::Redo(YsShellExtEdit *shl)
{
	YsShellExt *rawShl=(YsShellExt *)shl;
	rawShl->SetPolygonVertex(plHd,newVtHd.GetN(),newVtHd);
}

YSRESULT YsShellExtEdit::SetPolygonVertex(YsShellPolygonHandle plHd,YSSIZE_T nPlVt,const YsShellVertexHandle plVtHd[])
{
	YsShellExtEdit::UndoSetPolygonVertex *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new YsShellExtEdit::UndoSetPolygonVertex;
		undo->plHd=plHd;
		GetVertexListOfPolygon(undo->prvVtHd,plHd);
		undo->newVtHd.Set(nPlVt,plVtHd);
		AddUndo(undo);
	}

	return YsShellExt::SetPolygonVertex(plHd,nPlVt,plVtHd);
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoSetPolygonColor : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellPolygonHandle> plHdArray;
	YsArray <YsColor> prevCol,newCol;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoSetPolygonColor::Undo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		shl->SetColorOfPolygon(plHdArray[idx],prevCol[idx]);
	}
}

void YsShellExtEdit::UndoSetPolygonColor::Redo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		shl->SetColorOfPolygon(plHdArray[idx],newCol[idx]);
	}
}

YSRESULT YsShellExtEdit::SetPolygonColor(YsShellPolygonHandle plHd,const YsColor &col)
{
	YsShellExtEdit::UndoSetPolygonColor *undo=NULL;

	if(YSTRUE==takeEditLog)
	{
		undo=new YsShellExtEdit::UndoSetPolygonColor;

		YsColor prevCol;
		GetColorOfPolygon(prevCol,plHd);
		undo->plHdArray.MakeUnitLength(plHd);
		undo->prevCol.MakeUnitLength(prevCol);
		undo->newCol.MakeUnitLength(col);
	}

	YsShellExt::SetColorOfPolygon(plHd,col);

	if(NULL!=undo)
	{
		AddUndo(undo);
	}

	return YSOK;
}

YSRESULT YsShellExtEdit::SetPolygonAlpha(YsShellPolygonHandle plHd,const double alpha)
{
	YsShellExtEdit::UndoSetPolygonColor *undo=NULL;
	YsColor col;
	GetColorOfPolygon(col,plHd);

	if(YSTRUE==takeEditLog)
	{
		undo=new YsShellExtEdit::UndoSetPolygonColor;

		undo->plHdArray.MakeUnitLength(plHd);
		undo->prevCol.MakeUnitLength(col);
	}

	col.SetAd(alpha);
	YsShellExt::SetColorOfPolygon(plHd,col);

	if(NULL!=undo)
	{
		undo->newCol.MakeUnitLength(col);
		AddUndo(undo);
	}

	return YSOK;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoChangePolygonAttrib : public YsShellExtEdit::EditLog
{
public:
	YsShellPolygonHandle plHd;
	PolygonAttrib prvAttrib,newAttrib;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoChangePolygonAttrib::Undo(YsShellExtEdit *shl)
{
	PolygonAttrib *attr=shl->GetPolygonAttrib(plHd);
	attr->CopyFrom(prvAttrib);
}

void YsShellExtEdit::UndoChangePolygonAttrib::Redo(YsShellExtEdit *shl)
{
	PolygonAttrib *attr=shl->GetPolygonAttrib(plHd);
	attr->CopyFrom(newAttrib);
}

// YSRESULT YsShellExtEdit::SetPolygonSomething(YsShellPolygonHandle plHd,incoming prop)
// {
// 	PolygonAttrib *attr=GetPolygonAttrib(plHd);
// 	if(NULL!=attr)
// 	{
// 		UndoChangePolygonAttrib *undo=NULL;
// 		if(YSTRUE==takeEditLog)
// 		{
// 			undo=new UndoChangePolygonAttrib;
// 			undo->plHd=plHd;
// 			undo->prvAttrib.CopyFrom(*attr);
// 		}
// 
// 		attr->?=?;
// 
// 		if(NULL!=undo)
// 		{
// 			undo->newAttrib.CopyFrom(*attr);
// 		}
// 
// 		return YSOK;
// 	}
// 	return YSERR;
// }

YSRESULT YsShellExtEdit::SetPolygonDesiredElementSize(YsShellPolygonHandle plHd,const double elemSize)
{
	PolygonAttrib *attr=GetPolygonAttrib(plHd);
	if(NULL!=attr)
	{
		UndoChangePolygonAttrib *undo=NULL;
		if(YSTRUE==takeEditLog)
		{
			undo=new UndoChangePolygonAttrib;
			undo->plHd=plHd;
			undo->prvAttrib.CopyFrom(*attr);
		}

		attr->elemSize=elemSize;

		if(NULL!=undo)
		{
			undo->newAttrib.CopyFrom(*attr);
			AddUndo(undo);
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExtEdit::SetPolygonNoShadingFlag(YsShellPolygonHandle plHd,YSBOOL noShading)
{
	PolygonAttrib *attr=GetPolygonAttrib(plHd);
	if(NULL!=attr)
	{
		UndoChangePolygonAttrib *undo=NULL;
		if(YSTRUE==takeEditLog)
		{
			undo=new UndoChangePolygonAttrib;
			undo->plHd=plHd;
			undo->prvAttrib.CopyFrom(*attr);
		}

		attr->SetNoShading(noShading);

		if(NULL!=undo)
		{
			undo->newAttrib.CopyFrom(*attr);
			AddUndo(undo);
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExtEdit::SetPolygonTwoSidedFlag(YsShellPolygonHandle plHd,YSBOOL twoSided)
{
	PolygonAttrib *attr=GetPolygonAttrib(plHd);
	if(NULL!=attr)
	{
		UndoChangePolygonAttrib *undo=NULL;
		if(YSTRUE==takeEditLog)
		{
			undo=new UndoChangePolygonAttrib;
			undo->plHd=plHd;
			undo->prvAttrib.CopyFrom(*attr);
		}

		attr->SetTwoSided(twoSided);

		if(NULL!=undo)
		{
			undo->newAttrib.CopyFrom(*attr);
			AddUndo(undo);
		}

		return YSOK;
	}
	return YSERR;
}

////////////////////////////////////////////////////////////

YSRESULT YsShellExtEdit::SetPolygonAttrib(YsShellPolygonHandle plHd,const PolygonAttrib &inAttrib)
{
	if(YSOK==YsShellExt::SetPolygonAttrib(plHd,inAttrib))
	{
		UndoChangePolygonAttrib *undo=NULL;
		if(YSTRUE==takeEditLog)
		{
			undo=new UndoChangePolygonAttrib;
			undo->plHd=plHd;
			undo->prvAttrib=*GetPolygonAttrib(plHd);
			undo->newAttrib=inAttrib;
			AddUndo(undo);
		}
		return YSOK;
	}
	return YSERR;
}


////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoSetPolygonNormal : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellPolygonHandle> plHdArray;
	YsArray <YsVec3> prvNom,newNom;

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoSetPolygonNormal::Undo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		shl->SetNormalOfPolygon(plHdArray[idx],prvNom[idx]);
	}
}

void YsShellExtEdit::UndoSetPolygonNormal::Redo(YsShellExtEdit *shl)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		shl->SetNormalOfPolygon(plHdArray[idx],newNom[idx]);
	}
}

YSRESULT YsShellExtEdit::SetPolygonNormal(YsShellPolygonHandle plHd,const YsVec3 &nom)
{
	const YsShellPolygonHandle plHdArray[1]={plHd};
	const YsVec3 nomArray[1]={nom};
	return SetPolygonNormalMulti(1,plHdArray,nomArray);
}

YSRESULT YsShellExtEdit::SetPolygonNormalMulti(const YSSIZE_T nPl,const YsShellPolygonHandle plHdArray[],const YsVec3 nomArray[])
{
	UndoSetPolygonNormal *undo=NULL;

	if(YSTRUE==takeEditLog)
	{
		undo=new UndoSetPolygonNormal;
		undo->plHdArray.Set(nPl,plHdArray);
		undo->prvNom.Set(nPl,NULL);
		for(YSSIZE_T idx=0; idx<nPl; ++idx)
		{
			GetNormalOfPolygon(undo->prvNom[idx],plHdArray[idx]);
		}
		undo->newNom.Set(nPl,nomArray);
		AddUndo(undo);
	}

	for(YSSIZE_T idx=0; idx<nPl; ++idx)
	{
		SetNormalOfPolygon(plHdArray[idx],nomArray[idx]);
	}

	return YSOK;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoAddConstEdge : public YsShellExtEdit::EditLog
{
public:
	YsShellExt::ConstEdgeHandle ceHd;

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *shl);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoAddConstEdge::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=shl;
	shlExt->FreezeConstEdge(ceHd);
}

void YsShellExtEdit::UndoAddConstEdge::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=shl;
	shlExt->MeltConstEdge(ceHd);
}

void YsShellExtEdit::UndoAddConstEdge::NeverBeUndoneAgain(YsShellExtEdit *)
{
}

void YsShellExtEdit::UndoAddConstEdge::NeverBeRedoneAgain(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=shl;
	shlExt->DeleteFrozenConstEdge(ceHd);
}

YsShellExt::ConstEdgeHandle YsShellExtEdit::AddConstEdge(YSSIZE_T nVt,const YsShellVertexHandle ceVtHd[],YSBOOL isLoop)
{
	YsShellExt *shlExt=this;
	YsShellExt::ConstEdgeHandle ceHd=shlExt->AddConstEdge(nVt,ceVtHd,isLoop);

	if(YSTRUE==takeEditLog)
	{
		UndoAddConstEdge *undo=new UndoAddConstEdge;
		undo->ceHd=ceHd;
		AddUndo(undo);
	}

	return ceHd;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoModifyConstEdge : public YsShellExtEdit::EditLog
{
public:
	YsShellExtEdit::ConstEdgeHandle ceHd;
	YsArray <YsShellVertexHandle> prvCeVtHd,newCeVtHd;
	YSBOOL prvIsLoop,newIsLoop;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoModifyConstEdge::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->ModifyConstEdge(ceHd,prvCeVtHd,prvIsLoop);
}

void YsShellExtEdit::UndoModifyConstEdge::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->ModifyConstEdge(ceHd,newCeVtHd,newIsLoop);
}

YSRESULT YsShellExtEdit::ModifyConstEdge(ConstEdgeHandle ceHd,YSSIZE_T nVt,const YsShellVertexHandle ceVtHd[],YSBOOL isLoop)
{
	if(YSTFUNKNOWN==isLoop)
	{
		isLoop=IsConstEdgeLoop(ceHd);
	}

	if(YSTRUE==takeEditLog)
	{
		UndoModifyConstEdge *undo=new UndoModifyConstEdge;
		undo->ceHd=ceHd;
		GetConstEdge(undo->prvCeVtHd,undo->prvIsLoop,ceHd);
		undo->newCeVtHd.Set(nVt,ceVtHd);
		undo->newIsLoop=isLoop;
		AddUndo(undo);
	}

	return YsShellExt::ModifyConstEdge(ceHd,nVt,ceVtHd,isLoop);
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoDeleteMultiConstEdge : public YsShellExtEdit::EditLog
{
public:
	YsArray <YsShellExt::ConstEdgeHandle> ceHdArray;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *shl);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoDeleteMultiConstEdge::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	for(YSSIZE_T idx=0; idx<ceHdArray.GetN(); ++idx)
	{
		shlExt->MeltConstEdge(ceHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeleteMultiConstEdge::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	for(YSSIZE_T idx=0; idx<ceHdArray.GetN(); ++idx)
	{
		shlExt->FreezeConstEdge(ceHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeleteMultiConstEdge::NeverBeUndoneAgain(YsShellExtEdit *shl)
{
	// Means const edges will never be alive again.
	YsShellExt *shlExt=(YsShellExt *)shl;
	for(YSSIZE_T idx=0; idx<ceHdArray.GetN(); ++idx)
	{
		shlExt->DeleteFrozenConstEdge(ceHdArray[idx]);
	}
}

void YsShellExtEdit::UndoDeleteMultiConstEdge::NeverBeRedoneAgain(YsShellExtEdit *)
{
}

YSRESULT YsShellExtEdit::DeleteMultiConstEdge(YSSIZE_T nCe,const ConstEdgeHandle ceHdArray[])
{
	if(YSTRUE==takeEditLog)
	{
		UndoDeleteMultiConstEdge *undo=new UndoDeleteMultiConstEdge;
		undo->ceHdArray.Set(nCe,ceHdArray);
		AddUndo(undo);
	}

	for(YSSIZE_T idx=0; idx<nCe; ++idx)
	{
		YsShellExt::FreezeConstEdge(ceHdArray[idx]);
	}

	return YSOK;
}

YSRESULT YsShellExtEdit::DeleteConstEdge(ConstEdgeHandle ceHd)
{
	return DeleteMultiConstEdge(1,&ceHd);
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoSetConstEdgeAttrib : public YsShellExtEdit::EditLog
{
public:
	YsShellExtEdit::ConstEdgeHandle ceHd;
	YsShellExt::ConstEdge prvAttrib,newAttrib;
	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoSetConstEdgeAttrib::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->SetConstEdgeAttrib(ceHd,prvAttrib);
}

void YsShellExtEdit::UndoSetConstEdgeAttrib::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->SetConstEdgeAttrib(ceHd,newAttrib);
}

YSRESULT YsShellExtEdit::SetConstEdgeAttrib(YsShellExt::ConstEdgeHandle ceHd,const YsShellExt::ConstEdge &attrib)
{
	YsShellExt *shlExt=(YsShellExt *)this;

	UndoSetConstEdgeAttrib *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new UndoSetConstEdgeAttrib;
		undo->prvAttrib=shlExt->GetConstEdgeAttrib(ceHd);
		undo->newAttrib=attrib;
	}

	if(YSOK==shlExt->SetConstEdgeAttrib(ceHd,attrib))
	{
		if(NULL!=undo)
		{
			AddUndo(undo);
		}
		return YSOK;
	}
	else
	{
		if(NULL!=undo)
		{
			delete undo;
		}
		return YSERR;
	}
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoAddFaceGroup : public YsShellExtEdit::EditLog
{
public:
	YsShellExt::FaceGroupHandle fgHd;

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *shl);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoAddFaceGroup::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->FreezeFaceGroup(fgHd);
}

void YsShellExtEdit::UndoAddFaceGroup::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->MeltFaceGroup(fgHd);
}

void YsShellExtEdit::UndoAddFaceGroup::NeverBeUndoneAgain(YsShellExtEdit *)
{
}

void YsShellExtEdit::UndoAddFaceGroup::NeverBeRedoneAgain(YsShellExtEdit *shl)
{
	// Means never be melted again.
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->DeleteFrozenFaceGroup(fgHd);
}

YsShellExt::FaceGroupHandle YsShellExtEdit::AddFaceGroup(YSSIZE_T nPl,const YsShellPolygonHandle fgPlHd[])
{
	for(YSSIZE_T plIdx=0; plIdx<nPl; ++plIdx)
	{
		if(NULL!=FindFaceGroupFromPolygon(fgPlHd[plIdx]))
		{
			return NULL;  // A polygon already used.
		}
	}

	YsShellExt::FaceGroupHandle fgHd=YsShellExt::AddFaceGroup(nPl,fgPlHd);

	if(YSTRUE==takeEditLog)
	{
		UndoAddFaceGroup *undo=new UndoAddFaceGroup;
		undo->fgHd=fgHd;
		AddUndo(undo);
	}

	return fgHd;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoModifyFaceGroup : public YsShellExtEdit::EditLog
{
public:
	YsShellExt::FaceGroupHandle fgHd;
	YsArray <YsShellPolygonHandle> orgPlHd,newPlHd;

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoModifyFaceGroup::Undo(YsShellExtEdit *shl)
{
	YsShellExt &shlExt=(YsShellExt &)(*shl);
	shlExt.ModifyFaceGroup(fgHd,orgPlHd);
}

void YsShellExtEdit::UndoModifyFaceGroup::Redo(YsShellExtEdit *shl)
{
	YsShellExt &shlExt=(YsShellExt &)(*shl);
	shlExt.ModifyFaceGroup(fgHd,newPlHd);
}

YSRESULT YsShellExtEdit::AddFaceGroupPolygon(FaceGroupHandle fgHd,const YSSIZE_T nPl,const YsShellPolygonHandle fgPlHd[])
{
	UndoModifyFaceGroup *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new UndoModifyFaceGroup;
		undo->fgHd=fgHd;
		GetFaceGroup(undo->orgPlHd,fgHd);
	}

	if(YSOK==YsShellExt::AddFaceGroupPolygon(fgHd,nPl,fgPlHd))
	{
		if(NULL!=undo)
		{
			GetFaceGroup(undo->newPlHd,fgHd);
			AddUndo(undo);
		}
		return YSOK;
	}
	else
	{
		if(NULL!=undo)
		{
			delete undo;
		}
		return YSERR;
	}
}

YSRESULT YsShellExtEdit::ModifyFaceGroup(FaceGroupHandle fgHd,const YSSIZE_T nPl,const YsShellPolygonHandle fgPlHd[])
{
	UndoModifyFaceGroup *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new UndoModifyFaceGroup;
		undo->fgHd=fgHd;
		GetFaceGroup(undo->orgPlHd,fgHd);
		undo->newPlHd.Set(nPl,fgPlHd);
	}

	if(YSOK==YsShellExt::ModifyFaceGroup(fgHd,nPl,fgPlHd))
	{
		if(NULL!=undo)
		{
			AddUndo(undo);
		}
		return YSOK;
	}
	else
	{
		if(NULL!=undo)
		{
			delete undo;
		}
		return YSERR;
	}
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoDeleteFaceGroupPolygon : public YsShellExtEdit::EditLog
{
public:
	class FaceGroupPolygonPair
	{
	public:
		YsShellExt::FaceGroupHandle fgHd;
		YsShellPolygonHandle plHd;
	};

	YsArray <FaceGroupPolygonPair> fgPlPairArray;

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoDeleteFaceGroupPolygon::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	for(auto &fgPlPair : fgPlPairArray)
	{
		const YsShellPolygonHandle plHdArray[1]={fgPlPair.plHd};
		shlExt->AddFaceGroupPolygon(fgPlPair.fgHd,1,plHdArray);
	}
}

void YsShellExtEdit::UndoDeleteFaceGroupPolygon::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	for(auto &fgPlPair : fgPlPairArray)
	{
		shlExt->DeleteFaceGroupPolygon(fgPlPair.plHd);
	}
}

YsShellExt::FaceGroupHandle YsShellExtEdit::DeleteFaceGroupPolygon(YsShellPolygonHandle plHd)
{
	YsShellExt::FaceGroupHandle fgHd=YsShellExt::DeleteFaceGroupPolygon(plHd);
	if(NULL!=fgHd)
	{
		UndoDeleteFaceGroupPolygon *undo=NULL;
		if(YSTRUE==takeEditLog)
		{
			undo=new UndoDeleteFaceGroupPolygon;
			undo->fgPlPairArray.Set(1,NULL);
			undo->fgPlPairArray[0].fgHd=fgHd;
			undo->fgPlPairArray[0].plHd=plHd;
			AddUndo(undo);
		}
		return fgHd;
	}
	return NULL;
}

YSRESULT YsShellExtEdit::DeleteFaceGroupPolygonMulti(YSSIZE_T nPl,const YsShellPolygonHandle plHd[])
{
	YSRESULT res=YSERR;

	UndoDeleteFaceGroupPolygon *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new UndoDeleteFaceGroupPolygon;
	}

	for(YSSIZE_T plIdx=0; plIdx<nPl; ++plIdx)
	{
		YsShellExt::FaceGroupHandle fgHd=YsShellExt::DeleteFaceGroupPolygon(plHd[plIdx]);
		if(NULL!=fgHd)
		{
			res=YSOK;
			if(NULL!=undo)
			{
				undo->fgPlPairArray.Increment();
				undo->fgPlPairArray.Last().plHd=plHd[plIdx];
				undo->fgPlPairArray.Last().fgHd=fgHd;
			}
		}
	}

	if(NULL!=undo)
	{
		if(0<undo->fgPlPairArray.GetN())
		{
			AddUndo(undo);
		}
		else
		{
			delete undo;
		}
	}

	return res;
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoSetFaceGroupAttrib : public YsShellExtEdit::EditLog
{
public:
	YsShellExt::FaceGroupHandle fgHd;
	YsShellExt::FaceGroup prvAttrib,newAttrib;

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
};

void YsShellExtEdit::UndoSetFaceGroupAttrib::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->SetFaceGroupAttrib(fgHd,prvAttrib);
}

void YsShellExtEdit::UndoSetFaceGroupAttrib::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->SetFaceGroupAttrib(fgHd,newAttrib);
}

YSRESULT YsShellExtEdit::SetFaceGroupAttrib(YsShellExt::FaceGroupHandle fgHd,const YsShellExt::FaceGroup &attrib)
{
	YsShellExt *shlExt=(YsShellExt *)this;

	UndoSetFaceGroupAttrib *undo=NULL;
	if(YSTRUE==takeEditLog)
	{
		undo=new UndoSetFaceGroupAttrib;
		undo->prvAttrib=shlExt->GetFaceGroupAttrib(fgHd);
		undo->newAttrib=attrib;
	}

	if(YSOK==shlExt->SetFaceGroupAttrib(fgHd,attrib))
	{
		if(NULL!=undo)
		{
			AddUndo(undo);
		}
		return YSOK;
	}
	else
	{
		if(NULL!=undo)
		{
			delete undo;
		}
		return YSERR;
	}
}

////////////////////////////////////////////////////////////

class YsShellExtEdit::UndoDeleteFaceGroup : public YsShellExtEdit::EditLog
{
public:
	YsShellExt::FaceGroupHandle fgHd;

	virtual void Undo(YsShellExtEdit *shl);
	virtual void Redo(YsShellExtEdit *shl);
	virtual void NeverBeUndoneAgain(YsShellExtEdit *);
	virtual void NeverBeRedoneAgain(YsShellExtEdit *);
};

void YsShellExtEdit::UndoDeleteFaceGroup::Undo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->MeltFaceGroup(fgHd);
}

void YsShellExtEdit::UndoDeleteFaceGroup::Redo(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->FreezeFaceGroup(fgHd);
}

void YsShellExtEdit::UndoDeleteFaceGroup::NeverBeUndoneAgain(YsShellExtEdit *shl)
{
	YsShellExt *shlExt=(YsShellExt *)shl;
	shlExt->DeleteFrozenFaceGroup(fgHd);
}

void YsShellExtEdit::UndoDeleteFaceGroup::NeverBeRedoneAgain(YsShellExtEdit *)
{
}

YSRESULT YsShellExtEdit::DeleteFaceGroup(FaceGroupHandle fgHd)
{
	YsShellExt *shlExt=(YsShellExt *)this;

	if(YSOK==shlExt->FreezeFaceGroup(fgHd))
	{
		if(YSTRUE==takeEditLog)
		{
			UndoDeleteFaceGroup *undo=NULL;
			undo=new UndoDeleteFaceGroup;
			undo->fgHd=fgHd;
			AddUndo(undo);
		}
		return YSOK;
	}
	return YSERR;
}

